# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2026 Suturina Group

"""SimpNMR package entry point.

Exposes package version information.
"""

from .__version__ import __version__

__all__ = ["__version__"]
